/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009-2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend;

import java.util.concurrent.ConcurrentHashMap;

import javax.jbi.JBIException;

import org.ow2.petals.component.framework.se.AbstractServiceEngine;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.se.talend.model.TalendJobHandler;

/**
 * The petals-se-talend Service Engine.
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class TalendSe extends AbstractServiceEngine {

	/**
	 * This map is used to get the TalendJob associated with an end-point name.
	 * <p>This map is used every time the component receives a message.</p>
	 */
	private final ConcurrentHashMap<String,TalendJobHandler> endpointNameToJobHandler =
		new ConcurrentHashMap<String,TalendJobHandler> ();



	/* (non-Javadoc)
	 * @see org.ow2.petals.component.framework.AbstractComponent
	 * #doInit()
	 */
	@Override
	protected void doInit() throws JBIException {
		this.endpointNameToJobHandler.clear();
	}


	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.component.framework.se.AbstractServiceEngine
	 * #createServiceUnitManager()
	 */
	@Override
	protected AbstractServiceUnitManager createServiceUnitManager() {
		return new TalendSuManager( this );
	}


	/**
	 * Registers a Talend job and associates it with its end-point name.
	 * <p>
	 * When registering a new Talend job, its running status is set to false.
	 * </p>
	 * 
	 * @param talendJobHandler the Talend job handler to register (not null)
	 */
	public void registerTalendJobHandler( TalendJobHandler talendJobHandler ) {
		this.endpointNameToJobHandler.put( talendJobHandler.getEndpointName(), talendJobHandler );
	}


	/**
	 * Unregisters a Talend job handler.
	 * @param endpointName the end-point name (not null)
	 */
	public void unregisterTalendJobHandler( String endpointName ) {
		this.endpointNameToJobHandler.remove( endpointName );
	}


	/**
	 * Gets the Talend job associated with an end-point name.
	 * 
	 * @param endpointName an end-point name (not null)
	 * @return the Talend job handler behind this end-point, or null if none
	 */
	public TalendJobHandler getTalendJobHandler( String endpointName ) {
		return this.endpointNameToJobHandler.get( endpointName );
	}


	/**
	 * Determines whether the component is configured to validate all the exchanges with respect to the WSDLs.
	 * <p>
	 * If this method returns true, the service-unit configuration for the property
	 * "validate-exchange-by-wsdl" should be ignored.
	 * </p>
	 * 
	 * @return true if the exchange from all the SU should be validated with the WSDL, false otherwise
	 */
	public boolean validateExchangeByWsdl() {

		String s = getComponentExtensions().get( "validate-exchange-by-wsdl" );
		if( s != null )
			s = s.trim();

		return Boolean.valueOf( s );
	}
}
