/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend.interim;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description;
import org.ow2.easywsdl.schema.api.Import;
import org.ow2.easywsdl.schema.api.Schema;
import org.ow2.easywsdl.schema.api.SchemaWriter;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Operation;
import org.ow2.easywsdl.wsdl.api.Types;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.talend.utils.MepUtils;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * A utility class to validate exchanges with respect to a WSDL.
 * <p>
 * This class should not be used if no WSDL was provided.<br />
 * The validation being made is strict and only relies on the WSDL content.
 * </p>
 * <p>
 * The provided methods are thread-safe, since several objects may use it concurrently.
 * As an example, this class is currently used in {@link AbstractJBIListener}s.
 * </p>
 * <p>
 * This class is intended to be moved in the CDK.<br />
 * WSDL-based validation should not be enabled when attachments are expected
 * in the WSDL operation. base64binary is a simple type, but MTOM attachments
 * make data handlers have an element child (xop:include). The message validation
 * can only fail.
 * </p>
 * <p>
 * Hopefully, EasyWSDL will provide a better and in-memory solution.
 * </p>
 * 
 * @version 1.0
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class WsdlMessageChecker {

	/**
	 * The logger.
	 */
	private final Logger logger;

	/**
	 * The files containing the serialized XML schemas.
	 */
	private final Collection<File> schemaFiles;

	/**
	 * A XML Schema factory.
	 */
	private SchemaFactory jdkFactory;

	/**
	 * A map associating operation names and MEP.
	 * <p>
	 * Key = operation's name<br />
	 * Value = operation's MEP
	 * </p>
	 */
	private final Map<QName,MEPPatternConstants> operationNameToMep =
		new ConcurrentHashMap<QName,MEPPatternConstants> ();


	/**
	 * Constructor.
	 * @param wsdlLocation the WSDL location
	 * @param interfaceName the interface name whose validation properties must be cached
	 * @param logger the logger
	 * 
	 * @throws WSDLException if:
	 * <ul>
	 *  <li>The WSDL is invalid.</li>
	 *  <li>The WSDL does not contain the interface passed as an argument.</li>
	 *  <li>The WSDL interface does not contain any operation.</li>
	 * </ul>
	 */
	public WsdlMessageChecker( URL wsdlLocation, QName interfaceName, Logger logger )
	throws WSDLException {

		this.logger = logger;
		Description desc = UtilFactory.getWSDLUtil().createWsdlDescription( wsdlLocation, false );
		InterfaceType itf = desc.getInterface( interfaceName );

		if( itf != null ) {
			try {
				this.schemaFiles = storeWsdlSchemas( desc.getTypes());

			} catch( Exception e ) {
				throw new WSDLException( "The WSDL's schemas could not be cached for exchange validation." );
			}

			List<Operation> operations = itf.getOperations();
			if( operations == null || operations.isEmpty())
				throw new WSDLException( "The WSDL does not declare any operation." );

			for( Operation operation : itf.getOperations()) {
				QName operationName = operation.getQName();
				MEPPatternConstants mep = operation.getPattern();
				this.operationNameToMep.put( operationName, mep );
			}
		}
		else {
			throw new WSDLException( "The declared interface ("
					+ interfaceName.toString() + ") was not found in the provided WSDL." );
		}

		// Save memory
		desc = null;
	}


	/**
	 * Checks whether this exchange is valid with respect to the WSDL.
	 * <p>
	 * The following properties are checked:
	 * </p>
	 * <ul>
	 *  <li>The called operation is defined in the WSDL.</li>
	 *  <li>The used MEP is valid for this operation, as defined in the WSDL.</li>
	 *  <li>The input message is valid for this operation, as defined in the WSDL.</li>
	 * </ul>
	 * 
	 * <p>
	 * This method is thread-safe.
	 * </p>
	 * 
	 * @param exchange the exchange
	 * @return true if the exchange is valid, false otherwise
	 */
	public boolean validateExchange( Exchange exchange ) {

		QName operationName = exchange.getOperation();
		String exchangeId = exchange.getExchangeId();
		boolean result = false;

		try {
			result = validateOperation( operationName, exchangeId )
			&& validateMep( operationName, exchange.getPattern(), exchangeId )
			&& validateInputMessage(
					operationName,
					exchange.getInMessageContentAsDocument( true ), // Fork the source !
					exchangeId );

		} catch( MessagingException e ) {
			if( this.logger.isLoggable( Level.FINE ))
				this.logger.fine( "Exchange " + exchangeId + ": could not get the exchange content for WSDL-based validation." );
		}

		return result;
	}


	/**
	 * Checks whether this input message is valid with respect to the operation input schema.
	 * <p>
	 * This method is thread-safe.
	 * </p>
	 * 
	 * @param operationName the operation name
	 * @param inputMsg the input message
	 * @param exchangeId the exchange ID
	 * @return true if this message is valid with respect to the WSDL, false otherwise
	 * 
	 * <p>
	 * FIXME: this implementation makes disk accesses. Another (memory-based) implementation is expected in EasyWSDL.
	 * </p>
	 */
	public boolean validateInputMessage( QName operationName, Document inputMsg, String exchangeId ) {

		boolean result = false;

		// This is a temporary implementation
		// For every schema, we try to validate against it
		for( File f : this.schemaFiles) {

			// Validate the document using the JDK validation mechanism
			try {
				if( this.jdkFactory == null )
					this.jdkFactory = SchemaFactory.newInstance( XMLConstants.W3C_XML_SCHEMA_NS_URI );

				Validator jdkValidator = this.jdkFactory.newSchema( f ).newValidator();
				jdkValidator.validate( new DOMSource( inputMsg.getDocumentElement()));

				result = true;
				if( this.logger.isLoggable( Level.FINE ))
					this.logger.fine( "Exchange " + exchangeId + ": the input message was succesfully validated against the service's WSDL." );

				break;

			} catch( IOException e ) {
				if( this.logger.isLoggable( Level.FINE ))
					this.logger.fine( "Exchange " + exchangeId + ": failed to validate the input message against a cached schema." );

			} catch( SAXException e ) {
				if( this.logger.isLoggable( Level.FINE ))
					this.logger.fine( "Exchange " + exchangeId + ": failed to validate the input message against a cached schema." );
			}
		}

		if( ! result && this.logger.isLoggable( Level.FINE ))
			this.logger.finest( "Exchange " + exchangeId + ": invalid message for "
					+ operationName.toString() + "." );

		return result;
	}


	/**
	 * Checks whether this operation is associated with this MEP in the WSDL.
	 * <p>
	 * This method is thread-safe.
	 * </p>
	 * 
	 * @param operationName the operation name
	 * @param mep the MEP
	 * @param exchangeId the exchange ID
	 * @return true if this operation is associated with this MEP, false otherwise
	 */
	public boolean validateMep( QName operationName, URI mep, String exchangeId ) {

		MEPPatternConstants expectedMep = this.operationNameToMep.get( operationName );
		boolean result = MepUtils.compareMeps( expectedMep, mep );

		if( ! result && this.logger.isLoggable( Level.FINER )) {
			String msg = "Exchange " + exchangeId + ": invalid MEP for " + operationName.toString() + ".";
			msg += "\n" + expectedMep + " was expected instead of " + mep + ".";
			this.logger.finer( msg );
		}

		return result;
	}


	/**
	 * Checks whether this operation exists in the provided WSDL.
	 * <p>
	 * This method is thread-safe.
	 * </p>
	 * 
	 * @param operationName the operation name
	 * @param exchangeId the exchange ID
	 * @return true if the operation exists, false otherwise
	 */
	public boolean validateOperation( QName operationName, String exchangeId ) {

		boolean result = this.operationNameToMep.containsKey( operationName );
		if( ! result && this.logger.isLoggable( Level.FINE ))
			this.logger.fine( "Exchange " + exchangeId + ": invalid operation, "
					+ operationName.toString() + " was not found in the WSDL." );

		return result;
	}


	/**
	 * Serializes all the WSDL's schemas, updating the imports when necessary.
	 * 
	 * @param types the WSDL type element
	 * @return the list of schema files
	 * @throws Exception if something went wrong
	 */
	public static Collection<File> storeWsdlSchemas( Types types ) throws Exception {

		Map<String,File> nsToTargetFile = new HashMap<String,File> ();

		try {
			org.ow2.easywsdl.schema.SchemaFactory factory = org.ow2.easywsdl.schema.SchemaFactory.newInstance();
			SchemaWriter writer = factory.newSchemaWriter();

			// Associate every schema with a file
			for( Schema schema : types.getSchemas()) {
				String ns = schema.getTargetNamespace();
				File f = File.createTempFile( "petalsValidation_", ".xsd" );
				f.deleteOnExit();
				nsToTargetFile.put( ns, f );
			}

			// Update all the imports in the schema to reference the right file
			// And serialize it
			FileOutputStream fo = null;
			for( Schema schema : types.getSchemas()) {

				// Update the import locations
				for( Import imp_ : schema.getImports()) {
					File f = nsToTargetFile.get( imp_.getNamespaceURI());
					if( f != null ) {
						try {
							imp_.setLocationURI( new URI( f.getName()));

						} catch( URISyntaxException e ) {
							// should not happen
							e.printStackTrace();
							// otherwise, we won't miss it
						}
					} else
						throw new IOException( "The schema for (" + imp_.getNamespaceURI() + ") could not be resolved." );
				}

				// Write the schemas on the disk
				File f = nsToTargetFile.get( schema.getTargetNamespace());
				try {
					fo = new FileOutputStream( f, false );
					writer.writeSchema( schema, fo );

				} catch( Exception e ) {
					throw e;

				} finally {
					if( fo != null )
						fo.close();
				}
			}

		} catch( Exception e ) {

			for( File f : nsToTargetFile.values()) {
				if( ! f.delete())
					f.deleteOnExit();
			}

			throw e;
		}

		return nsToTargetFile.values();
	}


	/**
	 * Deletes the temporary files.
	 */
	public void deleteTamporaryFiles() {

		if( this.schemaFiles != null ) {
			for( File f : this.schemaFiles ) {
				if( f.exists() && ! f.delete())
					f.deleteOnExit();
			}
		}
	}
}
