/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009-2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend.model;

import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * A class to store information retrieved from the service-unit.
 * <p>
 * An instance of this class can only be created by the SU manager, when
 * a service-unit for the Talend SE is deployed. Undeployment of the same
 * service-unit should result in the destruction of this instance.
 * </p>
 * <p>
 * A service-unit can hold only one Talend job.
 * </p>
 * 
 * @version 1.0 of the internal model of the Talend SE
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class TalendJob {

    /**
     * Static integer used to generate identifiers.
     */
    private static AtomicInteger idCounter = new AtomicInteger();

    /**
     * Job field: identifier (generated, never null).
     */
    private final int id;

    /**
     * Job field: job name (provided in the SU configuration).
     * <p>
     * May be null.
     * </p>
     */
    private final String name;

    /**
     * Job field: job class name (provided in the SU configuration).
     * <p>
     * Cannot be null.
     * </p>
     */
    private final String className;

    /**
     * Job field: job context (provided in the SU configuration).
     * <p>
     * If a job is passed an incorrect context name, this is not dramatic.<br />
     * The job will use the default context in case where the given one is unknown.
     * </p>
     */
    private final String context;

    /**
     * The SU root path, to create a new class loader.
     */
    private final String suRootPath;

    /**
     * Job field: singleton (provided in SU configuration).
     * <p>
     * When a job is singleton, it means that at most one instance of this job can run at a given time.
     * <br />If a job is not singleton, it means we can have several instances of this job running at the same
     * moment.
     * </p>
     * <p>
     * This attribute should be used to determine whether an instance of this job can be created.
     * </p>
     */
    private final boolean singleton;

    /**
     * Container field: the end-point name (provided at deployment time).
     */
    private final String endpointName;

    /**
     * Container field: the service-unit name (provided at deployment time).
     */
    private final String serviceUnitName;

    /**
     * Computed field: the job's output attachments.
     * <p>
     * This list provides information to the component, so that it can query job context values,
     * load them and attach them in the returned exchange.
     * </p>
     */
    private final List<String> outAttachmentContextNames;



    /**
     * Constructor.
     * @param name the job name
     * @param className the job class name
     * @param context the job context
     * @param singleton true if the job is singleton, false otherwise
     * @param outAttachmentContextNames the attachments to retrieve and to forward when the job execution is over
     * @param endpointName the end-point name associated with this job
     * @param suRootPath the SU root path (to create the class loader)
     * @param serviceUnitName the service unit name
     */
    public TalendJob(
            String name,
            String className,
            String context,
            boolean singleton,
            List<String> outAttachmentContextNames,
            String endpointName,
            String suRootPath,
            String serviceUnitName ) {

        // Create the ID
        this.id = idCounter.getAndIncrement();

        // Store values
        this.name = name;
        this.className = className;
        this.context = context;
        this.singleton = singleton;
        this.outAttachmentContextNames = outAttachmentContextNames;
        this.endpointName = endpointName;
        this.serviceUnitName = serviceUnitName;
        this.suRootPath = suRootPath;
    }


    /**
     * @return the name
     */
    public String getName() {
        return this.name;
    }


    /**
     * @return the className
     */
    public String getClassName() {
        return this.className;
    }


    /**
     * @return the endpointName
     */
    public String getEndpointName() {
        return this.endpointName;
    }


    /**
     * @return the context
     */
    public String getContext() {
        return this.context;
    }


    /**
     * @return the singleton
     */
    public boolean isSingleton() {
        return this.singleton;
    }


    /**
     * @return the serviceUnitName
     */
    public String getServiceUnitName() {
        return this.serviceUnitName;
    }


    /**
     * @return the contextNameToAttachmentType
     */
    public List<String> getOutAttachmentContextNames() {
        return this.outAttachmentContextNames;
    }


    /**
     * @return the suRootPath
     */
    public String getSuRootPath() {
        return this.suRootPath;
    }


    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals( Object obj ) {

        return obj != null
        && obj instanceof TalendJob
        && ((TalendJob) obj).id == this.id;
    }


    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        return this.id;
    }
}
