/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend.utils;

import java.io.File;
import java.io.FileFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;

/**
 * A class loader that puts all the JARs and the root directory in the class path.
 * <p>
 * Copied from ClassLoaderUtils (CDK).
 * </p>
 * 
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class TalendClassLoaderUtils {

	/**
	 * As all methods of this class are static, no constructor is available.
	 */
	private TalendClassLoaderUtils() {
		// Nothing
	}

	/**
	 * Returns an URL[] table, with all Jar contained in the given directory path.
	 * @param directoryName directory containing the Jars
	 * @return an URL[] table
	 */
	public static URL[] getUrls( final String directoryName ) {

		final File path = new File( directoryName );
		final FileFilter jarFileFilter = new FileFilter() {
			public boolean accept( final File pathname ) {
				return pathname.getName().toLowerCase().endsWith( ".jar" );
			}
		};

		final File[] files = path.listFiles( jarFileFilter );
		final ArrayList<URL> urls = new ArrayList<URL>();
		String urlName = directoryName;
		if( ! urlName.endsWith( "/" ))
			urlName += "/";

		try {
			URL rootUrl = new URL( "file:///" + urlName );
			urls.add( rootUrl );

		} catch( MalformedURLException e1 ) {
			e1.printStackTrace();
			// Should never come here
		}


		if( files != null && files.length > 0 ) {
			for( final File file : files ) {
				try {
					if( file.isDirectory()) {
						urlName = file.getAbsolutePath();
						if( ! urlName.endsWith( "/" ))
							urlName += "/";

						URL rootUrl = new URL( "file:///" + urlName );
						urls.add( rootUrl );
					}
					else
						urls.add( file.toURI().toURL());


				} catch( final MalformedURLException e ) {
					// nothing
				}
			}
		}

		return urls.toArray( new URL[ urls.size()]);
	}

	/**
	 * Return an URLClassLoader object, with all Jar contained in the given directory path.
	 * 
	 * @param rootPath directory containing the Jars
	 * @param parent the parent class loader to use
	 * @return a classLoader
	 */
	public static URLClassLoader createClassLoader(final String rootPath, final ClassLoader parent) {
		return AccessController.doPrivileged( new TalendComponentPrivilegedAction( rootPath, parent ));
	}


	/**
	 * Create a new URL class loader.
	 */
	public static class TalendComponentPrivilegedAction implements PrivilegedAction<URLClassLoader> {

		/**
		 * The URLs to search classes.
		 */
		private URL[] urls = null;

		/**
		 * The parent class loader.
		 */
		private ClassLoader parent = null;

		/**
		 * Constructor.
		 * @param rootPath the root path
		 * @param parent the parent class loader
		 */
		public TalendComponentPrivilegedAction( final String rootPath, final ClassLoader parent ) {
			super();
			this.urls = getUrls( rootPath );
			this.parent = parent;
		}

		/*
		 * (non-Javadoc)
		 * @see java.security.PrivilegedAction#run()
		 */
		public URLClassLoader run() {
			return new TalendURLClassLoader( this.urls, this.parent );
		}
	}
}
