/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend.utils;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;

import org.w3c.dom.Node;

/**
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class TalendUtils {

	/**
	 * Deletes files recursively.
	 * @param files the files to delete
	 * @throws IOException if a file could not be deleted
	 */
	public static void deleteFilesRecursively( File... files ) throws IOException {

		if( files == null )
			return;

		for( File file : files ) {
            if (file.exists()) {

                if (file.isFile()) {
                    if (!file.delete())
                        throw new IOException(file.getAbsolutePath() + " could not be deleted.");

                } else if (file.isDirectory())
                    deleteFilesRecursively(file.listFiles());
            }
		}
	}


	/**
	 * Copies the content from in into os.
	 * <p>
	 * Neither <i>in</i> nor <i>os</i> are closed by this method.<br />
	 * They must be explicitly closed after this method is called.
	 * </p>
	 *
	 * @param in the input stream as the copy input
	 * @param os the output stream as the copy output
	 * @throws IOException if something went wrong
	 */
	public static void copyStream( InputStream in, OutputStream os ) throws IOException {

		byte[] buf = new byte[ 1024 ];
		int len;
		while((len = in.read( buf )) > 0) {
			os.write( buf, 0, len );
		}
	}


	/**
	 * Copies the content from in into outputFile.
	 * <p>
	 * <i>in</i> is not closed by this method.<br />
	 * It must be explicitly closed after this method is called.
	 * </p>
	 *
	 * @param in the input stream as the copy input
	 * @param outputFile will be created if it does not exist
	 * @throws IOException if something went wrong
	 */
	public static void copyStream( InputStream in, File outputFile ) throws IOException {

		if( ! outputFile.exists() && ! outputFile.createNewFile())
			throw new IOException( "Failed to create " + outputFile.getAbsolutePath() + "." );

		OutputStream os = new FileOutputStream( outputFile );
        try {
            copyStream(in, os);
        } finally {
            os.close();
        }
	}


	/**
	 * @param node a DOM node
	 * @return the name of the XML node (with no name space element)
	 */
	public static String getNodeName( Node node ) {

		String name = node.getNamespaceURI() != null ? node.getLocalName() : node.getNodeName();
		if( name.contains( ":" )) { //$NON-NLS-1$
			String[] parts = name.split( ":" ); //$NON-NLS-1$
			name = parts[ parts.length - 1 ];
		}

		return name;
	}


	/**
	 * Compares two strings, the first one being encoded as an URL and not the second.
	 * <p>
	 * This method is used to compare attachment names.<br />
	 * Method copied from the SOAP BC.
	 * </p>
	 *
	 * @param attachmentElementValue the value of the MTOM attachment, as referenced in the href attribute
	 * @param attachmentName the attachment name, as specified in the exchange
	 * @return true if the arguments are the same (despite the encoding), false otherwise
	 * @throws UnsupportedEncodingException if the encoding is not supported
	 */
	public static boolean compare( String attachmentElementValue, String attachmentName )
	throws UnsupportedEncodingException {

		if( attachmentElementValue.equals( attachmentName ))
			return true;

		// In some cases, the attachment ID is encoded as an URL
		String encodedText2 = URLEncoder.encode( attachmentName, "UTF-8" );
		if( attachmentElementValue.equals( encodedText2 ))
			return true;

		// Last attempt
		String encodedText1  = attachmentElementValue.replace( "@", "%40" );
		if( encodedText1.equals( encodedText2 ))
			return true;

		return false;
	}
}
